// Taken from http://www.eecg.toronto.edu/~amza/ece1747h/homeworks/examples/MPI/other-examples/mmult.c
// Modified by Tristan Aubrey-Jones 2013
// For comparsion with Flocc generated version
// see http://www.flocc.net/

/*
 *  mmult.c: matrix multiplication using MPI.
 * There are some simplifications here. The main one is that matrices B and C 
 * are fully allocated everywhere, even though only a portion of them is 
 * used by each processor (except for processor 0)
 */

#include <mpi.h>
#include <stdio.h>

#define SIZE 3000		/* Size of matrices */
#define N 3000

double A[SIZE][SIZE], B[SIZE][SIZE], C[SIZE][SIZE];

void fill_matrix(double m[SIZE][SIZE])
{
  static int n=0;
  int i, j;
  for (i=0; i<SIZE; i++)
    for (j=0; j<SIZE; j++)
      m[i][j] = 2.0;
}

void print_matrix(double m[SIZE][SIZE])
{
  int i, j = 0;
  for (i=0; i<SIZE; i++) {
    printf("\n\t| ");
    for (j=0; j<SIZE; j++)
      printf("%2f ", m[i][j]);
    printf("|");
  }
}


#define MAX_TIME_HEADER 1024
int timeRank;
char timeHeader[MAX_TIME_HEADER];
double beginTime, middleTime;

void initTime() {
  // rank
  timeRank = MPI::COMM_WORLD.Get_rank();
  int Nproc = MPI::COMM_WORLD.Get_size();

  // the current date/time
  time_t timer = time(NULL);
  struct tm* dateTime = localtime(&timer);
  char *strTime = asctime(dateTime);
  char* strPtr = strTime;
  while (*strPtr != '\0') { if (*strPtr == '\n') *strPtr = '\0'; strPtr++; }

  // make log header
  snprintf(timeHeader, MAX_TIME_HEADER, "%i: %s\t%i\t%i", timeRank, strTime, N, Nproc);
  
  // start time
  beginTime = MPI_Wtime();
  middleTime = beginTime;
}
 
  
void logTime(const char* event, double *startTime) {
  if (timeRank == 0) {
    double End = MPI_Wtime();
    std::cerr << timeHeader << "\t"
              << event << "\t" << (End - (*startTime)) << std::endl;
  }
  middleTime = MPI_Wtime();
} 


int main(int argc, char *argv[])
{
  int myrank, P, from, to, i, j, k;
  int tag = 666;		/* any value will do */
  MPI_Status status;
  double startTime, endTime;  

  MPI_Init (&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);	/* who am i */
  MPI_Comm_size(MPI_COMM_WORLD, &P); /* number of processors */

  /* Just to use the simple variants of MPI_Gather and MPI_Scatter we */
  /* impose that SIZE is divisible by P. By using the vector versions, */
  /* (MPI_Gatherv and MPI_Scatterv) it is easy to drop this restriction. */

  initTime();

  if (SIZE%P!=0) {
    if (myrank==0) printf("Matrix size not divisible by number of processors\n");
    MPI_Finalize();
   // exit(-1);
   return -1;
  }

  from = myrank * SIZE/P;
  to = (myrank+1) * SIZE/P;

  /* Process 0 fills the input matrices and broadcasts them to the rest */
  /* (actually, only the relevant stripe of A is sent to each process) */

  if (myrank==0) {
    fill_matrix(A);
    fill_matrix(B);
  }

  logTime("gendata", &beginTime);

  //if (myrank ==0) {
    startTime = MPI_Wtime();
  //}

  MPI_Bcast (B, SIZE*SIZE, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  MPI_Scatter (A, SIZE*SIZE/P, MPI_DOUBLE, A[from], SIZE*SIZE/P, MPI_DOUBLE, 0, MPI_COMM_WORLD);

  logTime("distdata", &beginTime);

  printf("computing slice %d (from row %d to %d)\n", myrank, from, to-1);
  for (i=from; i<to; i++) 
    for (j=0; j<SIZE; j++) {
      C[i][j]=0;
      for (k=0; k<SIZE; k++)
	C[i][j] += A[i][k]*B[k][j];
    }

  logTime("calcdata", &middleTime);

  //if (myrank ==0) {
    endTime = MPI_Wtime();
    printf("time on %i without gather is %f\n", P, endTime-startTime);
  //}

  MPI_Gather (C[from], SIZE*SIZE/P, MPI_DOUBLE, C, SIZE*SIZE/P, MPI_DOUBLE, 0, MPI_COMM_WORLD);

  //if (myrank ==0) {
    endTime = MPI_Wtime();
    printf("time on %i with gather is %f\n", P, endTime-startTime);
  //}
  /*if (myrank==0) {
    printf("\n\n");
    print_matrix(A);
    printf("\n\n\t       * \n");
    print_matrix(B);
    printf("\n\n\t       = \n");
    print_matrix(C);
    printf("\n\n");
  }*/

  MPI_Finalize();
  return 0;
}




